@@ -87,6 +87,7 @@ struct _GripGestureRegistration
   GPtrArray         *gesture_list;
   GeisInstance       instance;
   GIOChannel        *iochannel;
+  int                iochannel_id;
 };
 
 struct _GripRegistrationRequest
@@ -162,9 +163,9 @@ static void gesture_finish (void              *cookie,
                             GeisSize           attr_count,
                             GeisGestureAttr   *attrs);
 
-static void window_mapped_cb (GtkWidget       *widget,
-                              GdkEvent        *event,
-                              gpointer         user_data);
+static void toplevel_mapped_cb (GtkWidget       *toplevel,
+                                GdkEvent        *event,
+                                gpointer         user_data);
 
 static GeisGestureFuncs gesture_funcs = {
   gesture_added,
@@ -966,8 +967,15 @@ destroy_registration(GripGestureRegistration *reg)
     }
 
   g_list_free (reg->bindings);
-  g_io_channel_shutdown (reg->iochannel, TRUE, NULL);
+
+  g_io_channel_shutdown (reg->iochannel, FALSE, NULL);
+  g_source_remove (reg->iochannel_id);
+  g_io_channel_unref (reg->iochannel);
+
   geis_finish (reg->instance);
+
+  reg->instance = NULL;
+  reg->iochannel = NULL;
 }
 
 static void
@@ -1049,7 +1057,6 @@ new_window_registration(GripGestureManager *manager,
 {
   GripGestureRegistration *reg;
   GeisInstance  instance;
-  GIOChannel   *iochannel;
   gint fd = -1;
   GeisXcbWinInfo xcb_win_info = {
     .display_name = NULL,
@@ -1092,14 +1099,11 @@ new_window_registration(GripGestureManager *manager,
                     G_CALLBACK (window_destroyed_cb),
                     manager);
 
-  iochannel = g_io_channel_unix_new (fd);
-  g_io_add_watch (iochannel,
-                  G_IO_IN,
-                  io_callback,
-                  reg);
-
-  reg->iochannel = iochannel;
-
+  reg->iochannel = g_io_channel_unix_new (fd);
+  reg->iochannel_id = g_io_add_watch (reg->iochannel,
+                                      G_IO_IN,
+                                      io_callback,
+                                      reg);
   reg->gesture_list = g_ptr_array_new ();
 
   return reg;
@@ -1249,7 +1253,7 @@ toplevel_notify_cb (GtkWidget    *widget,
       GtkWidget *toplevel = gtk_widget_get_toplevel (widget);
       GripRegistrationRequest *req = (GripRegistrationRequest *)user_data;
 
-      if (GTK_IS_WINDOW (toplevel))
+      if (gtk_widget_is_toplevel (toplevel))
         {
           g_signal_handlers_disconnect_by_func (widget,
                                                 G_CALLBACK (toplevel_notify_cb),
@@ -1276,7 +1280,7 @@ toplevel_notify_cb (GtkWidget    *widget,
 
               g_signal_connect (toplevel,
                                 "map-event",
-                                G_CALLBACK (window_mapped_cb),
+                                G_CALLBACK (toplevel_mapped_cb),
                                 req->manager);
             }
         }
@@ -1312,7 +1316,7 @@ register_widget (GripGestureManager *manager,
 {
   GtkWidget *toplevel = gtk_widget_get_toplevel (widget);
 
-  if (GTK_IS_WINDOW (toplevel))
+  if (gtk_widget_is_toplevel (toplevel))
     {
       register_internal (manager,
                          widget,
@@ -1344,35 +1348,39 @@ register_widget (GripGestureManager *manager,
 }
 
 static void
-window_mapped_cb (GtkWidget   *widget,
-                  GdkEvent    *event G_GNUC_UNUSED,
-                  gpointer     user_data)
+toplevel_mapped_cb (GtkWidget   *toplevel,
+                    GdkEvent    *event G_GNUC_UNUSED,
+                    gpointer     user_data)
 {
   GripGestureManager *manager = (GripGestureManager *)user_data;
   GripGestureManagerPrivate *priv = manager->priv;
-  GList *tmp;
+  GList *request = priv->requests;
 
-  for (tmp = priv->requests; tmp != NULL; tmp = tmp->next)
+  while (request)
     {
-      GripRegistrationRequest *req = tmp->data;
+      GList *next = request->next;
+      GripRegistrationRequest *req = request->data;
 
-      register_widget (req->manager,
-                       req->widget,
-                       req->gesture_type,
-                       req->device_type,
-                       req->touch_points,
-                       req->callback,
-                       req->user_data,
-                       req->destroy);
+      if (gtk_widget_get_toplevel (req->widget) == toplevel)
+        {
+	  register_widget (req->manager,
+                           req->widget,
+                           req->gesture_type,
+                           req->device_type,
+                           req->touch_points,
+                           req->callback,
+                           req->user_data,
+                           req->destroy);
 
-      g_free (req);
+          priv->requests = g_list_remove_link (priv->requests, request);
+          g_free (req);
+          g_list_free_1 (request);
+        }
+      request = next;
     }
 
-  g_list_free (priv->requests);
-  priv->requests = NULL;
-
-  g_signal_handlers_disconnect_by_func (widget,
-                                        window_mapped_cb,
+  g_signal_handlers_disconnect_by_func (toplevel,
+                                        toplevel_mapped_cb,
                                         user_data);
 }
 
@@ -1441,7 +1449,7 @@ grip_gesture_manager_register_window (GripGestureManager  *manager,
 
           g_signal_connect (toplevel,
                             "map-event",
-                            G_CALLBACK (window_mapped_cb),
+                            G_CALLBACK (toplevel_mapped_cb),
                             manager);
         }
     }
