@@ -43,17 +43,25 @@ typedef enum {
 /**
  * SECTION:GripInputDevice
  *
- * The Grip input device represents input devices known to the OIF gesture
+ * The Grip input device represents input devices known to the uTouch gesture
  * recognition stack.
  */
 
-typedef struct AxisExtents
+typedef struct GripAxisExtents
 {
   gfloat minimum;
   gfloat maximum;
   gfloat resolution;
-} AxisExtents;
+} GripAxisExtents;
 
+#ifndef G_IR_IGNORE
+/**
+ * AxisExtents: (skip)
+ *
+ * A deprecated name for GripAxisExtents.
+ */
+typedef GripAxisExtents AxisExtents __attribute__((deprecated));
+#endif
 
 GType grip_input_device_get_type (void) G_GNUC_CONST;
 
@@ -107,10 +115,10 @@ grip_input_device_is_direct(GripInputDevice *input_device);
 gboolean
 grip_input_device_is_independent(GripInputDevice *input_device);
 
-AxisExtents *
+const GripAxisExtents *
 grip_input_device_get_x_extents(GripInputDevice *input_device);
 
-AxisExtents *
+const GripAxisExtents *
 grip_input_device_get_y_extents(GripInputDevice *input_device);
 
 
